CHIPSET(0x3577, I830_M,   "Intel(R) 830M")
CHIPSET(0x2562, 845_G,    "Intel(R) 845G")
CHIPSET(0x3582, I855_GM,  "Intel(R) 852GM/855GM")
CHIPSET(0x2572, I865_G,   "Intel(R) 865G")
CHIPSET(0x2582, I915_G,   "Intel(R) 915G")
CHIPSET(0x258A, E7221_G,  "Intel(R) E7221G (i915)")
CHIPSET(0x2592, I915_GM,  "Intel(R) 915GM")
CHIPSET(0x2772, I945_G,   "Intel(R) 945G")
CHIPSET(0x27A2, I945_GM,  "Intel(R) 945GM")
CHIPSET(0x27AE, I945_GME, "Intel(R) 945GME")
CHIPSET(0x29B2, Q35_G,    "Intel(R) Q35")
CHIPSET(0x29C2, G33_G,    "Intel(R) G33")
CHIPSET(0x29D2, Q33_G,    "Intel(R) Q33")
CHIPSET(0xA011, PNV_GM,   "Intel(R) Pineview M")
CHIPSET(0xA001, PNV_G,    "Intel(R) Pineview")
