/*
 * Minimal implementation of PanoramiX/Xinerama
 */

#ifndef XSERVER_PSEUDORAMIX_H
#define XSERVER_PSEUDORAMIX_H

void
PseudoramiXAddScreen(int x, int y, int w, int h);
void
PseudoramiXResetScreens(void);

#endif /* XSERVER_PSEUDORAMIX_H */
